unit Dbxtnav;
{
extendet navigator and related components
to use one navigator for all mdi childs.
the navigator will attach itself to the datasource
of the active dbcontrol.
You must place a DBXT component on each form to
use the navigator with.
}

interface

uses
  Forms,classes,controls,ExtCtrls,windows,messages,
  buttons,typinfo,sysutils,dialogs,
  db,dbtables,dbitypes,dbctrls,
  dbxloc,dbxbm1,dbxbm2,sercheng;

  const nversion = '2.51';

Type

  {Navigator and buttons}
  TXNavButton = class;
  TNavDataLink = class;
  TXDBNavigator = class;

  TXNavGlyph = (ngEnabled, ngDisabled);
  TXNavigateBtn = (nbFirst, nbPrior, nbNext, nbLast,
                  nbInsert, nbDelete, nbEdit, nbPost, nbCancel, nbRefresh,
                  nbSearch, nbSearchNext, nbSetBookmark, nbGotoBookmark );

  TXButtonSet = set of TXNavigateBtn;
  TXNavButtonStyle = set of (nsAllowTimer, nsFocusRect);
  TButtonSet = set of TXNavigateBtn;

  ENavClick = procedure (Sender: TObject; Button: TXNavigateBtn) of object;

  {locate options}
  TSearchOptions = set of (soExact, soWildcards, soCaseSensitive);
  TSearchDirection = (sdUp,sdDown,sdBeginning);

  {some events}
  TDBXTGetDSEvent = function(Sender: TObject; var ds: TDatasource): boolean of object;
  TXDBNavButtonEvent = procedure(Sender: TObject;btn: array of boolean) of object;

  {used to save the search conditions}
  TXPR=class
    private
      field: TField;
      cs:boolean;    {case sensitive}
      dir: integer;  {direction 0- down, 1 - up, 2- from now}
      wild: boolean; {use wildcards}
      ex:boolean;    {exact match}
    public
      constructor create(f: TField;c: boolean;d:integer;w:boolean;x:boolean);
  end;

  {dbextender component must be on each form to connect
  with the dbxnavigator}
  TDBXT = class(TComponent)
  private
    { Private-Deklarationen }
    fSearchOptions:TSearchOptions;
    fSearchDirection:TSearchDirection;
    fSaveOptions:boolean;
    ffieldonly: boolean;    {use only active field for search}
    fonlyvisible:boolean;   {use only visible fields for search}
    fMasteronly: boolean;   {use only masterdatasource, dont check for datasource
                             of current db control}
    NavList: TList;         {list of attached navigators
                            (i never used more than one :-) )}
    fDataSet: TDBDataset;   {current dataset}
    fDatasource: TDatasource; {current datasource}
    fField: string;         {current field}
    fListSort:boolean;      {sort field list in locate form combo ?}
    oldactivate: TNotifyevent;  {save hooked events}
    olddeactivate: TNotifyevent;
    oldctrlchange: TNotifyevent;
    xprdslist: TStringList; {used with locate}
    xprlist: TList;         {used with locate}
    bmdslist: TStringList;   {list with bookmark descriptions}
    bmlist: TList;           {list with bookmarks}
    FMastersource:TDatasource;
    FOnGetDS: TDBXTGetDSEvent;
    sresult:boolean;  {search result}
    inactive:boolean; {used in destroy}
    {Set / Get Mastersource}
    function GetMasterSource: TDataSource;
    procedure SetMasterSource(Value: TDataSource);
    {called in activate / deactivate Events of parent form}
    procedure parentformactivate(sender: TObject);
    procedure parentformdeactivate(sender: TObject);
    {called when active control changed}
    procedure ctrlchanged(sender: TObject);
    procedure GetNavigators;
    procedure SetXT;
    procedure ClearXT;
    function GetListSort:boolean;
    procedure SetListSort(b:boolean);
  protected
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    { Protected-Deklarationen }
  public
    { Public-Deklarationen }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    {search in field s for p,
     search could be casesensitive (cs), use wildcards (wc),
     up/down/from now (dir) and exact match (x)}
    procedure UserSearch(s,p:string;cs,wc:boolean;dir:word;x:boolean);
    {call search dialog and search}
    function Search:boolean;
    {search next after search or usersearch}
    procedure Searchnext;
    {get result of last search}
    function GetSearchResult:boolean;
    {ask if searchnext is possible (ie. search or usersearch was
    called before)}
    function cansearchnext:boolean;

    procedure SetBookmark;
    procedure GotoBookmark;
    function hasbookmark:boolean;
    {in case i forgot something this will re-set tdbxt to
    use the current datasource}
    procedure refresh;
  published
  { Published-Deklarationen }
    property VisibleFields: boolean read fonlyvisible write fonlyvisible;
    property FieldOnly: boolean read ffieldonly write ffieldonly;
    property MasterOnly: boolean read fMasteronly write fMasteronly;
    property MasterSource: TDataSource read GetMasterSource write SetMasterSource;
    property OnGetDatasource: TDBXTGetDSEvent read FOnGetDS write FOnGetDS;
    property SortedList: boolean read getlistsort write setlistsort;
    property SearchOptions: TSearchOptions read fSearchOptions write fSearchOptions;
    property SearchDirection: TSearchDirection read fSearchDirection write fSearchDirection;
    property SaveOptions:boolean read fSaveOptions write fSaveOptions;
    {use this to change OnActivate/ OnDeactivate of parent form}
    property onParentActivate: TNotifyevent read oldactivate write oldactivate;
    property onParentDeActivate: TNotifyevent read olddeactivate write olddeactivate;
    property onScreenActiveControlChange: TNotifyevent read oldctrlchange write oldctrlchange;
  end;

{ TDBNavigator }

  TXDBNavigator = class (TCustomPanel)
  private
    fGetRecCount:boolean;{get RecCount ? (sucks on sqlserver)}
    frecnum:longint;     {record number}
    freccount:longint;   {record count}
    FVisibleButtons: TButtonSet;
    FDataLink : TNavDataLink;
    FdefaultDBXT: TDBXT; {use this dbxt only}
    FDBXT: TDBXT;        {current dbxt}
    FHints: TStrings;
    ButtonWidth: Integer;
    fSpacing1: Integer;
    fSpacing2: Integer;
    MinBtnSize: TPoint;
    FOnNavClick: ENavClick;
    FOnChange: TNotifyEvent;
    FOnButtonChange: TXDBNavButtonEvent;
    FOnSearchFail: TNotifyEvent;
    FocusedButton: TXNavigateBtn;
    FConfirmDelete: Boolean;
    function GetdefDBXT: TDBXT;
    procedure SetdefDBXT(Value: TDBXT);
    function GetDBXT: TDBXT;
    procedure SetDBXT(Value: TDBXT);
    function GetDataSource: TDataSource;
    procedure SetDataSource(Value: TDataSource; XT: TDBXT);
    procedure GetRecNum;
    procedure InitButtons;
    procedure InitHints;
    procedure Click(Sender: TObject);
    procedure BtnMouseDown (Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SetVisible(Value: TButtonSet);
    procedure AdjustSize (var W: Integer; var H: Integer);
    procedure SetHints(Value: TStrings);
    procedure WMSize(var Message: TWMSize);  message WM_SIZE;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure SearchChanged(value: boolean; XT: TDBXT);
    procedure BookmarkChanged(value: boolean; XT: TDBXT);
  protected
    Buttons: array[TXNavigateBtn] of TXNavButton;
    procedure DataChanged;
    procedure EditingChanged;
    procedure ActiveChanged;
    procedure SomethingChanged;
    procedure Loaded; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
    property XT: TDBXT read GetDBXT write SetDBXT;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    {click navbutton}
    procedure BtnClick(Index: TXNavigateBtn);
    {current datasource}
    property DataSource: TDataSource read GetDataSource;
    {recno}
    property recordnumber: longint read frecnum;
    {reccount}
    property recordcount: longint read freccount;
  published
    {use only default xt when set}
    property defaultXT: TDBXT read GetdefDBXT write SetdefDBXT;
    property VisibleButtons: TButtonSet read FVisibleButtons write SetVisible
      default [nbFirst, nbPrior, nbNext, nbLast, nbInsert, nbDelete,
        nbEdit, nbPost, nbCancel, nbRefresh, nbSearch, nbSearchNext,
        nbSetBookmark, nbGotoBookmark];
    property Align;
    property DragCursor;
    property DragMode;
    property Enabled;
    property GetRecCount:boolean read fGetRecCount write fGetRecCount;
    property Ctl3D;
    property Hints: TStrings read FHints write SetHints;
    property ParentCtl3D;
    property ParentShowHint;
    property PopupMenu;
    property ConfirmDelete: Boolean read FConfirmDelete write FConfirmDelete default True;
    {space between buttons}
    property Spacing: Integer read FSpacing1 write FSpacing1;
    {space between buttongroups}
    property GroupSpacing1: Integer read FSpacing2 write FSpacing2;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnClick: ENavClick read FOnNavClick write FOnNavClick;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnSearchFail: TNotifyEvent read FOnSearchFail write FOnSearchFail;
    {somethings changed (enabled / disabled, no bookmark etc}
    property OnButtonChange: TXDBNavButtonEvent read FOnButtonChange write FOnButtonChange;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnStartDrag;
    property OnEnter;
    property OnExit;
    property OnResize;
  end;

{ TNavButton }

  TXNavButton = class(TSpeedButton)
  private
    FIndex: TXNavigateBtn;
    FNavStyle: TXNavButtonStyle;
    FRepeatTimer: TTimer;
    procedure TimerExpired(Sender: TObject);
  protected
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    destructor Destroy; override;
    property NavStyle: TXNavButtonStyle read FNavStyle write FNavStyle;
    property Index : TXNavigateBtn read FIndex write FIndex;
  end;

  { TNavDataLink }

  TNavDataLink = class(TDataLink)
  private
    FNavigator: TXDBNavigator;
  protected
    procedure EditingChanged; override;
    procedure DataSetChanged; override;
    procedure ActiveChanged; override;
  public
    constructor Create(ANav: TXDBNavigator);
    destructor Destroy; override;
  end;



procedure Register;
{get the datasource from active control}
procedure GetDSFromCtrl(ctl: TWinControl;var ds: TDatasource; var dT: TDBDataset;
                       var df: string);


var
  LocateForm: TLocateForm; {search}
  SetBMForm: TSetBMForm;   {set bookmark}
  GotoBMForm: TGotoBMForm; {goto bookmark}

implementation

uses dbconsts,dbxconst;

const
  BtnTypeName: array[TXNavigateBtn] of PChar = ('FIRST', 'PRIOR', 'NEXT',
    'LAST', 'INSERT', 'DELETE', 'EDIT', 'POST', 'CANCEL', 'REFRESH', 'SEARCH',
    'SEARCHNEXT','SETBOOKMARK','GOTOBOOKMARK');

{$IFDEF VER120} {Borland Delphi 4.0 }
  BtnHintId: array[TXNavigateBtn] of Pointer = (@SFirstRecord, @SPriorRecord,
    @SNextRecord, @SLastRecord, @SInsertRecord, @SDeleteRecord, @SEditRecord,
    @SPostEdit, @SCancelEdit, @SRefreshRecord, @sSearch, @sSearchNext, @sSetBookMark,
    @sGotoBookMark);

{$ELSE}
  BtnHintId: array[TXNavigateBtn] of Word = (SFirstRecord, SPriorRecord,
    SNextRecord, SLastRecord, SInsertRecord, SDeleteRecord, SEditRecord,
    SPostEdit, SCancelEdit, SRefreshRecord, sSearch, sSearchNext, sSetBookMark,
    sGotoBookMark);

{$ENDIF}

procedure Register;
begin
  RegisterComponents('Datensteuerung', [TDBXT,TXDBNavigator]);
end;



{the simple way ...}
{
procedure GetDSFromCtrl(ctl: TWinControl;var ds: TDatasource; var dT: TDBDataset;
                       var df: string);
var
   fm: TForm;
   cmp: TComponent;
begin
  ds:=nil;
  dt:=nil;
  df:='';
  if ctl = nil then exit;
  if ctl is TDBedit then
  begin
   try
     with ctl as TDBEdit do
     begin
          ds:=datasource;
          df:=datafield;
     end;
   except
         ds:=nil;
   end;
  end;
  if ctl is TDBGrid then
  begin
   try
     with ctl as TDBGrid do
     begin
          ds:=datasource;
     end;
     except
           ds:=nil;
     end;
  end;
  if ctl is TDBMemo then
  begin
   try
     with ctl as TDBMemo do
     begin
          ds:=datasource;
          df:=datafield;
     end;
     except
           ds:=nil;
     end;
  end;
  if ctl is TDBListBox then
  begin
   try
     with ctl as TDBListBox do
     begin
          ds:=datasource;
          df:=datafield;
     end;
     except
           ds:=nil;
     end;
  end;
  if ctl is TDBComboBox then
  begin
   try
     with ctl as TDBComboBox do
     begin
          ds:=datasource;
          df:=datafield;
     end;
     except
           ds:=nil;
     end;
  end;
  if ctl is TDBImage then
  begin
   try
     with ctl as TDBImage do
     begin
          ds:=datasource;
          df:=datafield;
     end;
     except
           ds:=nil;
     end;
  end;
  if ctl is TDBCheckbox then
  begin
   try
     with ctl as TDBCheckbox do
     begin
          ds:=datasource;
          df:=datafield;
     end;
     except
           ds:=nil;
     end;
  end;
  if ctl is TDBRadiogroup then
  begin
   try
     with ctl as TDBRadiogroup do
     begin
          ds:=datasource;
          df:=datafield;
     end;
     except
           ds:=nil;
     end;
  end;
  if ctl is TDBLookUpCombo then
  begin
   try
     with ctl as TDBLookUpCombo do
     begin
          ds:=datasource;
          df:=datafield;
     end;
     except
           ds:=nil;
     end;
  end;
  if ctl is TDBLookUpList then
  begin
   try
     with ctl as TDBLookUpList do
     begin
          ds:=datasource;
          df:=datafield;
     end;
     except
           ds:=nil;
     end;
  end;
  if ctl is TDBDateEdit then
  begin
   try
     with ctl as TDBDateEdit do
     begin
          ds:=datasource;
          df:=datafield;
     end;
     except
           ds:=nil;
     end;
  end;
  if ds <> nil then
     dt:=ds.dataset as TDBDataset
  else
     dt:=nil;
end;
}

{... and the better way}
procedure GetDSFromCtrl(ctl: TWinControl;var ds: TDatasource; var dT: TDBDataset;
                       var df: string);
var
  PropInfo: PPropInfo;
begin
   ds:=nil;
   df:='';
   dt:=nil;
   if ctl = nil then exit;
   try
     PropInfo := TypInfo.GetPropInfo(ctl.ClassInfo,'Datasource');
     if (PropInfo <> nil) and (PropInfo^.PropType^.Kind = tkClass) then
       ds := TObject(GetOrdProp(ctl, PropInfo)) as TDataSource
   except
     ds:=nil;
   end;
   if ds <> nil then
   begin
     dt:=ds.dataset as TDBDataset;
     try
       PropInfo := TypInfo.GetPropInfo(ctl.ClassInfo,'DataField');
       if (PropInfo <> nil) and (PropInfo^.PropType^.Kind = tkString) then
         df := GetStrProp(ctl, PropInfo);
       if (PropInfo <> nil) and (PropInfo^.PropType^.Kind = tkLString) then
         df := GetStrProp(ctl, PropInfo);

      except
       df:='';
      end;
    end;
end;


{call stringcomparefunctions}
function xCompareStrings(Target, Pattern: PChar;cs,wild,x:boolean):boolean;
begin
 if x then {exact match}
 begin
   if cs then result:=StrComp(target,pattern) = 0
   else result:=StrIComp(target,pattern) = 0
 end
 else
   if wild then
	 	result:=CompareStrings(target,pattern,cs,wild)
   else
   if cs then result:=StrLComp(target,pattern,StrLen(pattern)) = 0
   else result:=StrLIComp(target,pattern,StrLen(pattern)) = 0
end;



constructor TXPR.create(f: TField;c: boolean;d:integer;w:boolean;x:boolean);
begin
   field:=f;
   cs:=c;
   dir:=d;
   wild:=w;
   ex:=x;
end;



procedure TDBXT.parentformactivate(sender: TObject);
begin
 {form with dbxt gets active}
 if not (csDesigning in ComponentState) then
 begin
     oldctrlchange:=screen.onactivecontrolchange;
     screen.onactivecontrolchange:=ctrlchanged;
     GetNavigators; {look for navigators, fill list}
     SetXT;         {inform navigators about self}
     if assigned(oldactivate) then oldactivate(sender);
 end;
end;

procedure TDBXT.parentformdeactivate(sender: TObject);
begin
 if not (csDesigning in ComponentState) then
   screen.onactivecontrolchange:=oldctrlchange;
 oldctrlchange:=nil;
 ClearXT;  {remove dbxt from navigators}
 NavList.clear;  {clear navigators}
 if assigned(olddeactivate) then olddeactivate(sender);
end;

constructor TDBXT.Create(AOwner: TComponent);
begin
     fonlyvisible:=true;
     ffieldonly:=false;
     fdatasource:=nil;
     sresult:=false;
     fListSort:=false;
     fSearchOptions:=[];
     fSearchDirection:=sdBeginning;
     fSaveOptions:=true;
     xprdslist:=TStringList.create;
     xprList:=TList.create;
     bmdslist:=TStringList.create;
     bmList:=TList.create;
     NavList:=TList.create;
     inactive:=true;
     if not (csDesigning in ComponentState) then
     begin
     with aowner as TForm do
     begin
          {never change OnActivate or OnDeactivate of
           the parentform after creation!, use OnParent... instead}
          oldactivate:=onactivate;
          olddeactivate:=ondeactivate;
          oldctrlchange:=nil;
          ondeactivate:=parentformdeactivate;
          onactivate:=parentformactivate;
     end;
     end;
     inherited create(aowner);
end;

destructor TDBXT.Destroy;
begin
     if not (csDesigning in ComponentState) then
     {if form is closed, onactivecontrolchange
     is already cleared in parentformdeactivate}
     if not inactive then
       with owner as TForm do
            screen.onactivecontrolchange:=oldctrlchange;
     xprdslist.free;
     xprlist.free;
     bmdslist.free;
     bmlist.free;
     {navlist, onActivate / .. ?}
     inherited destroy;
end;

function TDBXT.GetListSort:boolean;
begin
     result:=fListSort;
end;

procedure TDBXT.SetListSort(b:boolean);
begin
     fListSort:=b;
end;

procedure TDBXT.GetNavigators;
var
   n: integer;
begin
  with owner as TForm do
  begin
    for n:=0 to componentcount-1 do
    begin
      if components[n] is TXDBNavigator then
         NavList.Add(components[n]);
    end;
  end;
  if application.mainform.formstyle = fsMDIForm then
  begin
  with application.mainform do
  begin
    for n:=0 to componentcount-1 do
    begin
      if components[n] is TXDBNavigator then
         NavList.Add(components[n]);
    end;
  end;
  end;
end;

procedure TDBXT.SetXT;
var
   n: integer;
begin
   inactive:=false;
   for n:=0 to NavList.count-1 do
       TXDBNavigator(NavList[n]).SetDBXT(Self);
end;

procedure TDBXT.ClearXT;
var
   n: integer;
begin
   inactive:=true;
   for n:=0 to NavList.count-1 do
       TXDBNavigator(NavList[n]).SetDBXT(nil);
end;

procedure TDBXT.refresh;
var
   n:integer;
begin
  if fDatasource = nil then begin
     fDatasource:=fMasterSource;
     if fDatasource <> nil then
          fDataset:=fDatasource.dataset as TDBDataset
     else
         fDataset:=nil;
  end;
  for n:=0 to NavList.count-1 do
     TXDBNavigator(NavList[n]).setdatasource(FDatasource,self);
end;

procedure TDBXT.ctrlchanged;
var
   c: TWinControl;
   n: integer;
begin
  c:=screen.activecontrol;
  if c = nil then exit;
  fField:='';
  GetDSFromCtrl(c,FDatasource,fDataset,fField);
  if fmasteronly then fdatasource:=nil;
  if fDatasource = nil then begin
     fDatasource:=fMasterSource;
     if fDatasource <> nil then
          fDataset:=fDatasource.dataset as TDBDataset
     else
         fDataset:=nil;
  end;
  for n:=0 to NavList.count-1 do
     TXDBNavigator(NavList[n]).setdatasource(FDatasource,self);
  if assigned(oldctrlchange) then oldctrlchange(sender);
end;



procedure TDBXT.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
end;

procedure TDBXT.UserSearch(s,p:string;cs,wc:boolean;dir:word;x:boolean);
var
   found: boolean;
   f: TField;
   n: integer;
   bookmark: TBookmark;
   searchexpr: string;
   r:boolean;
   Target, Pattern: Array[0..255] of Char;
begin
     sresult:=false;
     if fdataset = nil then exit;
     if dir > 2 then exit;
     bookmark:=fdataset.GetBookMark;
     f:=fdataset.fieldbyname(s);
     if f = nil then exit;
     searchexpr:=p;
     found:=false;
     StrPCopy(Pattern,searchexpr);
     case  dir of
     0: begin
        try
        fdataset.disablecontrols;
        while not fdataset.eof do
        begin
            StrPCopy(Target,f.asstring);
            r:=xCompareStrings(target,pattern,cs,wc,x);
            if r then
            begin
                  found:=true;
                  break;
             end;
             fdataset.next;
        end;
        finally
              fdataset.enablecontrols;
        end;
        end;
     1: begin
        try
        fdataset.disablecontrols;
        while not fdataset.bof do
        begin
            StrPCopy(Target,f.asstring);
            r:=xCompareStrings(target,pattern,cs,wc,x);
             if r  then
             begin
                  found:=true;
                  break;
             end;
             fdataset.prior;
        end;
        finally
              fdataset.enablecontrols;
        end;
        end;
     2: begin
             try
        fdataset.disablecontrols;
        fdataset.first;
        while not fdataset.eof do
        begin
            StrPCopy(Target,f.asstring);
            r:=xCompareStrings(target,pattern,cs,wc,x);
             if r then
             begin
                  found:=true;
                  break;
             end;
             fdataset.next;
        end;
        finally
              fdataset.enablecontrols;
        end;
        end;
     end;
     n:=xprdslist.indexofobject(fdataset);
     {add this datasource to the list to
      do a searchnext with this params}
     if n <= -1 then
     begin
          xprdslist.addobject(searchexpr,fdataset);
          xprlist.add(TXPR.Create(f,cs,dir,wc,x));
     end
     else
     begin
          xprdslist.delete(n);
          xprlist.delete(n);
          xprdslist.addobject(searchexpr,fdataset);
          xprlist.add(TXPR.Create(f,cs,dir,wc,x));
     end;
     {notify navigators, searchnext button must be enabled}
     for n:=0 to NavList.Count-1 do
       TXDBNavigator(NavList[n]).SearchChanged(true,self);
     if not found then fdataset.gotobookmark(bookmark);
     fdataset.freebookmark(bookmark);
     sresult:=found; {for getsearchresult}
end;


function TDBXT.Search:boolean;
var
   found: boolean;
   f: TField;
   n: integer;
   bookmark: TBookmark;
   searchexpr: string;
   r:boolean;
   res:integer;
   Target, Pattern: Array[0..255] of Char;
   exactmatch,usewildcards:boolean;
   casesensitive:boolean;
   searchdirection:integer;
   datefield: boolean;
   datetext:string;
   dateformat:string;
   locopt: TLocateOptions;

    function canlocate:boolean;
    begin
      {delphi 2 does not build the expression tree of the filter
      correct for boolean fields, so locate fails.
      (the filter for a boolean field named TEST is simply
      TEST, not TEST=TRUE}
      result:=(not usewildcards) and (f.datatype <> ftBoolean) and
      not f.calculated;
    end;


 begin
     datefield:=false;
     LocateForm := TLocateForm.Create( self );
     Locateform.hlabel.hint:='DBXTools version '+nversion;
     LocateForm.table:=fDataset;
     LocateForm.fieldonly:=ffieldonly;
     LocateForm.onlyvisible:=fonlyvisible;
     LocateForm.combobox1.sorted:=fListSort;
     LocateForm.StartField:=fField;

     n:=xprdslist.indexofobject(fdataset);
     if (n <= -1) or not fSaveOptions then
     begin {no search before}
     LocateForm.checkbox1.checked:=soCaseSensitive in fSearchOptions;
     LocateForm.checkbox2.checked:=soWildCards in fSearchOptions;
     LocateForm.checkbox3.checked:=soExact in fSearchOptions;
     LocateForm.radiogroup1.itemindex:=cardinal(fSearchDirection);


     end
     else
     begin
     LocateForm.checkbox1.checked:=TXPR(xprlist[n]).cs;
     LocateForm.checkbox2.checked:=TXPR(xprlist[n]).wild;
     LocateForm.checkbox3.checked:=TXPR(xprlist[n]).ex;
     LocateForm.radiogroup1.itemindex:=TXPR(xprlist[n]).dir;
     end;

     result:=false;
     res:=LocateForm.Showmodal;

     if res = mrOK then
     begin
     result:=true;
     if fdataset = nil then exit;
     bookmark:=fdataset.GetBookMark;
     f:=fdataset.fieldbyname(Locateform.SearchField);
     searchexpr:=LocateForm.edit1.text;
     found:=false;
     casesensitive:=LocateForm.checkbox1.checked;
     usewildcards:=LocateForm.checkbox2.checked;
     exactmatch:=LocateForm.checkbox3.checked;
     searchdirection:=LocateForm.radiogroup1.itemindex;

     {special treatment of date for string compare}

     if f is TDateTimeField then
     begin
       datefield:=true;
       if (f as TDateTimeField).DisplayFormat <> '' then
          dateFormat := (f as TDateTimeField).DisplayFormat
       else
       case f.DataType of
        ftDate: dateFormat := ShortDateFormat;
        ftTime: dateFormat := LongTimeFormat;
       end;
       try
         DateTimeToString(searchexpr, dateFormat, StrToDate(searchexpr));
       except
         searchexpr:=LocateForm.edit1.text;
       end;
     end;


     StrPCopy(Pattern,searchexpr);
     case  searchdirection of
     0: begin {down}
        try
        fdataset.disablecontrols;
        while not fdataset.eof do
        begin
            if datefield then
            begin
               DateTimeToString(dateText, dateFormat, f.asDateTime);
               StrPCopy(Target,datetext);
            end
            else
               StrPCopy(Target,f.asstring);
               r:=xCompareStrings(target,pattern,casesensitive,usewildcards,exactmatch);
            if r then
            begin
                  found:=true;
                  break;
            end;
            fdataset.next;
        end;
        finally
              fdataset.enablecontrols;
        end;
        end;
     1: begin {up}
        try
        fdataset.disablecontrols;
        while not fdataset.bof do
        begin
            if datefield then
            begin
               DateTimeToString(dateText, dateFormat, f.asDateTime);
               StrPCopy(Target,datetext);
            end
            else
               StrPCopy(Target,f.asstring);
            r:=xCompareStrings(target,pattern,casesensitive,usewildcards,exactmatch);
             if r  then
             begin
                  found:=true;
                  break;
             end;
             fdataset.prior;
        end;
        finally
              fdataset.enablecontrols;
        end;
        end;
     2: begin {from start}
        try
        fdataset.disablecontrols;

        if not canlocate then
        begin

        fdataset.first;
        while not fdataset.eof do
        begin
            if datefield then
            begin
               DateTimeToString(dateText, dateFormat, f.asDateTime);
               StrPCopy(Target,datetext);
            end
            else
               StrPCopy(Target,f.asstring);
            r:=xCompareStrings(target,pattern,casesensitive,usewildcards,exactmatch);
             if r then
             begin
                  found:=true;
                  break;
             end;
             fdataset.next;

        end;
        end
        else
        begin

             locopt:=[];
             if not exactmatch then
               include(locopt,loPartialkey);
             if not casesensitive then
               include(locopt,loCaseInsensitive);
             found:=fdataset.locate(f.fieldname,searchexpr,locopt);
        end



        finally
              fdataset.enablecontrols;
        end;


        end;
     end;
     {add this datasource to the list to
      do a searchnext with this params}
     n:=xprdslist.indexofobject(fdataset);
     if n <= -1 then
     begin

          xprdslist.addobject(searchexpr,fdataset);
          xprlist.add(TXPR.Create(f,casesensitive,searchdirection,
                      usewildcards,exactmatch));

     end
     else
     begin
          xprdslist.delete(n);
          xprlist.delete(n);
          xprdslist.addobject(searchexpr,fdataset);
          xprlist.add(TXPR.Create(f,casesensitive,searchdirection,
                      usewildcards,exactmatch));
     end;
     {notify navigators, searchnext button must be enabled}
     for n:=0 to NavList.Count-1 do
       TXDBNavigator(NavList[n]).SearchChanged(true,self);
     if not found then fdataset.gotobookmark(bookmark);
     fdataset.freebookmark(bookmark);
     sresult:=found; {for getsearchresult}
     end;
     LocateForm.free;
end;

procedure TDBXT.SearchNext;
var
   found: boolean;
   f: TField;
   n: integer;
   bookmark: TBookmark;
   searchexpr: string;
   x,r,wild,cs:boolean;
   dir:integer;
   Target, Pattern: Array[0..255] of Char;
begin
     if fdataset = nil then exit;
     if fdataset.state <> dsBrowse then exit;
     n:=xprdslist.indexofobject(fdataset);
     if n <= -1 then exit; {no search before}
     bookmark:=fdataset.GetBookMark;
     f:=TXPR(xprlist[n]).field;
     searchexpr:=xprdslist[n];
     StrPCopy(Pattern,searchexpr);
     found:=false;
     cs:=TXPR(xprlist[n]).cs;
     dir:=TXPR(xprlist[n]).dir;
     wild:=TXPR(xprlist[n]).wild;
     x:=TXPR(xprlist[n]).ex;
     if dir=2 then dir:=0;
     case  dir of
     0: begin
        try
        fdataset.disablecontrols;
        fdataset.next;
        while not fdataset.eof do
        begin
            StrPCopy(Target,f.asstring);
            r:=xCompareStrings(target,pattern,cs,wild,x);
             if r  then
             begin
                  found:=true;
                  break;
             end;
             fdataset.next;
        end;
        finally
              fdataset.enablecontrols;
        end;
        end;
     1: begin
        try
        fdataset.disablecontrols;
        fdataset.prior;
        while not fdataset.bof do
        begin
            StrPCopy(Target,f.asstring);
            r:=xCompareStrings(target,pattern,cs,wild,x);
             if r  then
             begin
                  found:=true;
                  break;
             end;
             fdataset.prior;
        end;
        finally
              fdataset.enablecontrols;
        end;
        end;
     end;
     if not found then fdataset.gotobookmark(bookmark);
     fdataset.freebookmark(bookmark);
     sresult:=found;
end;


function TDBXT.CanSearchNext:boolean;
begin
   if fdatasource = nil then result:=false
   else
   result:=(xprdslist.indexofobject(fdataset) > -1);
end;

function TDBXT.GetSearchResult:boolean;
begin
   result:=sresult;
end;

procedure TDBXT.SetBookmark;
var
   bookmark: TBookmark;
   n:integer;
   res: integer;
begin
     if fdataset = nil then exit;
     SetBMForm := TSetBMForm.Create( self );
     SetBMForm.hlabel.hint:='DBXTools version '+nversion;
     res:=SetBMForm.Showmodal;
     if res = mrOK then
     begin
        bookmark:=fdataset.GetBookMark;
        bmdslist.addobject(SetBMForm.edit1.text,fdataset);
        bmlist.add(bookmark);
        {inform Navigators, bookmark button changes}
        for n:=0 to NavList.Count-1 do
        TXDBNavigator(NavList[n]).BookMarkChanged(true,self);
     end;
     SetBMForm.Free;
end;

procedure TDBXT.GotoBookmark;
var
   n:integer;
   res: integer;
   bookmark: TBookmark;
begin
     if fdataset = nil then exit;
     GotoBMForm := TGotoBMForm.Create( self );
     GotoBMForm.hlabel.hint:='DBXTools version '+nversion;
     GotoBMForm.listbox1.items:=bmdslist;
     res:=GotoBMForm.Showmodal;
     if res = mrOK then
     begin
          n:=GotoBMForm.listbox1.itemindex;
          if n >= 0 then
          begin
            bookmark:=tbookmark(bmlist[n]);
            fdataset.GotoBookMark(bookmark);
          end;
     end
     else if res = mrAbort then
     begin
          n:=GotoBMForm.listbox1.itemindex;
          if n >= 0 then
          begin
            bookmark:=tbookmark(bmlist[n]);
            fdataset.FreeBookMark(bookmark);
            bmlist.delete(n);
            bmdslist.delete(n);
            if not hasbookmark then
            for n:=0 to NavList.Count-1 do
             TXDBNavigator(NavList[n]).BookMarkChanged(false,self);
          end;
     end;
     GotoBMForm.Free;
end;


function TDBXT.HasBookmark:boolean;
begin
   if fdatasource = nil then result:=false
   else
   result:=(bmdslist.indexofobject(fdataset) > -1);
end;

function TDBXT.GetMasterSource: TDataSource;
var
   xds: TDatasource;
   xres: boolean;
begin
  xres:=false;
  xds:=nil;
  if Assigned(FOnGetDS) then xres:=FOnGetDS(Self,xds);
  if xres then
  Result := xds
  else
  Result := FMasterSource;
end;

procedure TDBXT.SetMasterSource(Value: TDataSource);
var
   n:integer;
begin
  FMasterSource := Value;

  if fmasteronly or (fdatasource = nil) then
  begin
     fDatasource:=fMasterSource;
     if fDatasource <> nil then
          fDataset:=fDatasource.dataset as TDBDataset
     else
         fDataset:=nil;
     fField:='';
  end;
  for n:=0 to NavList.count-1 do
     TXDBNavigator(NavList[n]).setdatasource(FDatasource,self);
end;


constructor TXDBNavigator.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csAcceptsControls, csSetCaption] +
    [csFramed, csOpaque];
  FDataLink := TNavDataLink.Create(Self);

  FVisibleButtons := [nbFirst, nbPrior, nbNext, nbLast, nbInsert,
    nbDelete, nbEdit, nbPost, nbCancel, nbRefresh, nbSearch, nbSearchNext,
    nbSetBookmark, nbGotoBookmark];

  FHints := TStringList.Create;
  InitButtons;
  BevelOuter := bvNone;
  BevelInner := bvNone;
  Width := 241;
  Height := 25;
  ButtonWidth := 0;
  FocusedButton := nbFirst;
  FConfirmDelete := True;
  fSpacing1:=2;
  fSpacing2:=4;
  fRecnum := -1;
  freccount:=-1;
  fGetRecCount:=true;
end;

destructor TXDBNavigator.Destroy;
begin
  FDataLink.Free;
  FDataLink := nil;
  FHints.free;
  inherited Destroy;
end;


procedure TXDBNavigator.InitButtons;
var
  I: TXNavigateBtn;
  Btn: TXNavButton;
  X: Integer;
  ResName: array[0..40] of Char;
begin
  MinBtnSize := Point(20, 18);
  X := 0;
  for I := Low(Buttons) to High(Buttons) do
  begin
    Btn := TXNavButton.Create (Self);
    Btn.Index := I;
    Btn.Visible := I in FVisibleButtons;
    Btn.Enabled := false;
    Btn.SetBounds (X, 0, MinBtnSize.X, MinBtnSize.Y);
    Btn.Glyph.Handle := LoadBitmap(HInstance,
        StrFmt(ResName, 'dbn_%s', [BtnTypeName[I]]));
    Btn.NumGlyphs := 2;
    Btn.OnClick := Click;
    Btn.OnMouseDown := BtnMouseDown;
    Btn.Parent := Self;
    Buttons[I] := Btn;
    X := X + MinBtnSize.X+4;
  end;
  InitHints;
  Buttons[nbPrior].NavStyle := Buttons[nbPrior].NavStyle + [nsAllowTimer];
  Buttons[nbNext].NavStyle  := Buttons[nbNext].NavStyle + [nsAllowTimer];
end;

procedure TXDBNavigator.InitHints;
var
  I: Integer;
  J: TXNavigateBtn;
begin
  for J := Low(Buttons) to High(Buttons) do
{$IFDEF VER120} {Borland Delphi 4.0 }
    Buttons[J].Hint := LoadResString(BtnHintId[J]);
{$ELSE}
    Buttons[J].Hint := LoadStr (BtnHintId[J]);
{$ENDIF}
  J := Low(Buttons);
  for I := 0 to (FHints.Count - 1) do
  begin
    if FHints.Strings[I] <> '' then Buttons[J].Hint := FHints.Strings[I];
    if J = High(Buttons) then Exit;
    Inc(J);
  end;
end;

procedure TXDBNavigator.SetHints(Value: TStrings);
begin
  FHints.Assign(Value);
  InitHints;
end;

procedure TXDBNavigator.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FDataLink <> nil) and
    (AComponent = DataSource) then
  begin
      FDataLink.DataSource := nil;
      if not (csLoading in ComponentState) then
        ActiveChanged;
  end;
end;

procedure TXDBNavigator.SetVisible(Value: TButtonSet);
var
  I: TXNavigateBtn;
  W, H: Integer;
begin
  W := Width;
  H := Height;
  FVisibleButtons := Value;
  for I := Low(Buttons) to High(Buttons) do
    Buttons[I].Visible := I in FVisibleButtons;
  AdjustSize (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds (Left, Top, W, H);
  Invalidate;
end;

procedure TXDBNavigator.AdjustSize (var W: Integer; var H: Integer);
var
  Count: Integer;
  MinW: Integer;
  I: TXNavigateBtn;
  Space, Temp, Remain: Integer;
  X: Integer;
  aspace1: integer;
begin
  if (csLoading in ComponentState) then Exit;
  if Buttons[nbFirst] = nil then Exit;

  Count := 0;
  for I := Low(Buttons) to High(Buttons) do
  begin
    if Buttons[I].Visible then
    begin
      Inc(Count);
    end;
  end;
  if Count = 0 then Inc(Count);

  MinW := Count * (MinBtnSize.X - 1) +((count-1) * fspacing1) + 1;
  aspace1:=0;
  if (nbSearch in FVisibleButtons) or (nbSearchnext in FVisibleButtons) then
  begin
    inc(minw,fspacing2);
    aspace1:=fSpacing2;
  end;
  if W < MinW then
    W := MinW;
  if H < MinBtnSize.Y then
    H := MinBtnSize.Y;

  ButtonWidth := ((W - 1 -((count-1) * fspacing1)-aspace1) div Count) + 1;
  Temp := Count * (ButtonWidth - 1) + 1 + ((count-1) * fspacing1)+aspace1;
  if Align = alNone then
    W := Temp;

  X := 0;
  Remain := W - Temp;
  Temp := Count div 2;
  for I := Low(Buttons) to High(Buttons) do
  begin
    if Buttons[I].Visible then
    begin
      Space := 0;
      if i > nbrefresh then
      begin
           inc(x,aspace1);
           aspace1:=0;
      end;
      if Remain <> 0 then
      begin
        Dec (Temp, Remain);
        if Temp < 0 then
        begin
          Inc (Temp, Count);
          Space := 1;
        end;
      end;
      Buttons[I].SetBounds (X, 0, ButtonWidth + Space, Height);
{      Inc (X, ButtonWidth - 1 + Space+fSpacing1);}
      Inc (X, ButtonWidth - 1 + Space+fspacing1);
    end
    else
      Buttons[I].SetBounds (Width + 1, 0, ButtonWidth, Height);
  end;
end;

procedure TXDBNavigator.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  W, H: Integer;
begin
  W := AWidth;
  H := AHeight;
  AdjustSize (W, H);
  inherited SetBounds (ALeft, ATop, W, H);
end;

procedure TXDBNavigator.WMSize(var Message: TWMSize);
var
  W, H: Integer;
begin
  inherited;

  { check for minimum size }
  W := Width;
  H := Height;
  AdjustSize (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds(Left, Top, W, H);
  Message.Result := 0;
end;

procedure TXDBNavigator.Click(Sender: TObject);
begin
  BtnClick (TXNavButton (Sender).Index);
end;

procedure TXDBNavigator.BtnMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  OldFocus: TXNavigateBtn;
begin
  OldFocus := FocusedButton;
  FocusedButton := TXNavButton (Sender).Index;
  if TabStop and (GetFocus <> Handle) and CanFocus then
  begin
    SetFocus;
    if (GetFocus <> Handle) then
      Exit;
  end
  else if TabStop and (GetFocus = Handle) and (OldFocus <> FocusedButton) then
  begin
    Buttons[OldFocus].Invalidate;
    Buttons[FocusedButton].Invalidate;
  end;
end;

procedure TXDBNavigator.BtnClick(Index: TXNavigateBtn);
begin
  if (DataSource <> nil) and (DataSource.State <> dsInactive) then
  begin
    with DataSource.DataSet do
    begin
      case Index of
        nbPrior: Prior;
        nbNext: Next;
        nbFirst: First;
        nbLast: Last;
        nbRefresh: Refresh;
        nbInsert: Insert;
        nbEdit: Edit;
        nbCancel: Cancel;
        nbPost: Post;
        nbDelete:
          begin
{$IFDEF VER120} {Borland Delphi 4.0 }
             if not FConfirmDelete or
                (MessageDlg (SDeleteRecordQuestion,
                mtConfirmation, mbOKCancel, 0) <> idCancel) then
{$ELSE}
            if not FConfirmDelete or
                (MessageDlg (LoadStr(SDeleteRecordQuestion),
                mtConfirmation, mbOKCancel, 0) <> idCancel) then
{$ENDIF}
              Delete;
          end;
        nbSearch:
          begin
             if XT.Search then
             if assigned(FOnSearchFail) and not XT.GetSearchResult  then
                FOnSearchFail(self);
          end;
        nbSearchNext:
          begin
             XT.SearchNext;
             if assigned(FOnSearchFail) and not XT.GetSearchResult  then
                FOnSearchFail(self);
          end;
        nbSetBookmark:XT.SetBookmark;
        nbGotoBookmark: XT.GotoBookmark;
      end;
    end;
  end;
  if not (csDesigning in ComponentState) and Assigned(FOnNavClick) then
    FOnNavClick(Self, Index);
end;

procedure TXDBNavigator.WMSetFocus(var Message: TWMSetFocus);
begin
  Buttons[FocusedButton].Invalidate;
end;

procedure TXDBNavigator.WMKillFocus(var Message: TWMKillFocus);
begin
  Buttons[FocusedButton].Invalidate;
end;

procedure TXDBNavigator.KeyDown(var Key: Word; Shift: TShiftState);
var
  NewFocus: TXNavigateBtn;
  OldFocus: TXNavigateBtn;
begin
  OldFocus := FocusedButton;
  case Key of
    VK_RIGHT:
      begin
        NewFocus := FocusedButton;
        repeat
          if NewFocus < High(Buttons) then
            NewFocus := Succ(NewFocus);
        until (NewFocus = High(Buttons)) or (Buttons[NewFocus].Visible);
        if NewFocus <> FocusedButton then
        begin
          FocusedButton := NewFocus;
          Buttons[OldFocus].Invalidate;
          Buttons[FocusedButton].Invalidate;
        end;
      end;
    VK_LEFT:
      begin
        NewFocus := FocusedButton;
        repeat
          if NewFocus > Low(Buttons) then
            NewFocus := Pred(NewFocus);
        until (NewFocus = Low(Buttons)) or (Buttons[NewFocus].Visible);
        if NewFocus <> FocusedButton then
        begin
          FocusedButton := NewFocus;
          Buttons[OldFocus].Invalidate;
          Buttons[FocusedButton].Invalidate;
        end;
      end;
    VK_SPACE:
      begin
        if Buttons[FocusedButton].Enabled then
          Buttons[FocusedButton].Click;
      end;
  end;
end;

procedure TXDBNavigator.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTARROWS;
end;

procedure TXDBNavigator.DataChanged;
begin
  SomeThingChanged;
{  GetRecNum;}
{  if Assigned(FOnChange) then FOnChange(Self);}
  if Assigned(FOnButtonChange) then FOnButtonChange(Self,
  [Buttons[nbFirst].Enabled,Buttons[nbPrior].Enabled, Buttons[nbNext].Enabled,
  Buttons[nbLast].Enabled,Buttons[nbInsert].Enabled,Buttons[nbDelete].Enabled,Buttons[nbEdit].Enabled,
  Buttons[nbPost].Enabled,Buttons[nbCancel].Enabled,Buttons[nbRefresh].Enabled,
  Buttons[nbSearch].Enabled,Buttons[nbSearchNext].Enabled,
  Buttons[nbSetBookmark].Enabled,Buttons[nbGotoBookmark].Enabled]);
end;

procedure TXDBNavigator.GetRecNum;
var
  cP: CurProps;
  rP: RECProps;
begin

  fRecnum := -1;
  freccount:=-1;
  if fdatalink.datasource <> nil then
  if fdatalink.datasource.dataset <> nil then
  with FDataLink.DataSource.dataset do
  begin
    try
      if state = dsInactive then
      begin
         fRecnum := -1;
         freccount:=-1;
         exit;
      end;
      {$IFDEF WIN32}
       fRecnum := recno;
       if fGetRecCount then
          freccount:=RecordCount
       else
          freccount:=-1;
      {$ELSE}
      {we need to make this call to grab the cursor's iSeqNums}
      Check(DBiGetCursorProps(Handle,cP));
      {synchronize the BDE cursor with the dataset's cursor}
      UpdateCursorPos;
      {fill rP with the current record's properties}
      Check(DBiGetRecord(Handle,DBiNOLOCK,nil,@rP));
      {what kind of dataset are we looking at?}
      case cP.iSeqNums of
        0: fRecnum := rP.iPhyRecNum; {dBase}
        1: fRecnum := rP.iSeqNum;    {Paradox}
      end;
      if fGetReccount then
         freccount:=RecordCount {errr}
      else
         freccount:=-1;
      {$ENDIF}
    except
      fRecnum := -1;
      freccount:=-1;
    end;
  end;
end;

procedure TXDBNavigator.SearchChanged(value: boolean; XT: TDBXT);
begin
  if ((fdefaultDBXT <> nil) and (fdefaultDBXT = XT)) or (fdefaultDBXT = nil) then
    Buttons[nbSearchNext].Enabled := value;
  if Assigned(FOnButtonChange) then FOnButtonChange(Self,
  [Buttons[nbFirst].Enabled,Buttons[nbPrior].Enabled, Buttons[nbNext].Enabled,
  Buttons[nbLast].Enabled,Buttons[nbInsert].Enabled,Buttons[nbDelete].Enabled,Buttons[nbEdit].Enabled,
  Buttons[nbPost].Enabled,Buttons[nbCancel].Enabled,Buttons[nbRefresh].Enabled,
  Buttons[nbSearch].Enabled,Buttons[nbSearchNext].Enabled,
  Buttons[nbSetBookmark].Enabled,Buttons[nbGotoBookmark].Enabled]);
end;

procedure TXDBNavigator.BookmarkChanged(value: boolean; XT: TDBXT);
begin
  if ((fdefaultDBXT <> nil) and (fdefaultDBXT = XT)) or (fdefaultDBXT = nil) then
    Buttons[nbGotoBookmark].Enabled := value;
  if Assigned(FOnButtonChange) then FOnButtonChange(Self,
  [Buttons[nbFirst].Enabled,Buttons[nbPrior].Enabled, Buttons[nbNext].Enabled,
  Buttons[nbLast].Enabled,Buttons[nbInsert].Enabled,Buttons[nbDelete].Enabled,Buttons[nbEdit].Enabled,
  Buttons[nbPost].Enabled,Buttons[nbCancel].Enabled,Buttons[nbRefresh].Enabled,
  Buttons[nbSearch].Enabled,Buttons[nbSearchNext].Enabled,
  Buttons[nbSetBookmark].Enabled,Buttons[nbGotoBookmark].Enabled]);
end;

procedure TXDBNavigator.EditingChanged;
begin
  SomeThingChanged;
  if Assigned(FOnButtonChange) then FOnButtonChange(Self,
  [Buttons[nbFirst].Enabled,Buttons[nbPrior].Enabled, Buttons[nbNext].Enabled,
  Buttons[nbLast].Enabled,Buttons[nbInsert].Enabled,Buttons[nbDelete].Enabled,Buttons[nbEdit].Enabled,
  Buttons[nbPost].Enabled,Buttons[nbCancel].Enabled,Buttons[nbRefresh].Enabled,
  Buttons[nbSearch].Enabled,Buttons[nbSearchNext].Enabled,
  Buttons[nbSetBookmark].Enabled,Buttons[nbGotoBookmark].Enabled]);
end;

procedure TXDBNavigator.ActiveChanged;
var
  I: TXNavigateBtn;
begin
  if not (Enabled and FDataLink.Active) then
  begin
    for I := Low(Buttons) to High(Buttons) do
      Buttons[I].Enabled := False;
    GetRecNum;
    if Assigned(FOnButtonChange) then FOnButtonChange(Self,
    [Buttons[nbFirst].Enabled,Buttons[nbPrior].Enabled, Buttons[nbNext].Enabled,
    Buttons[nbLast].Enabled,Buttons[nbInsert].Enabled,Buttons[nbDelete].Enabled,Buttons[nbEdit].Enabled,
    Buttons[nbPost].Enabled,Buttons[nbCancel].Enabled,Buttons[nbRefresh].Enabled,
    Buttons[nbSearch].Enabled,Buttons[nbSearchNext].Enabled,
    Buttons[nbSetBookmark].Enabled,Buttons[nbGotoBookmark].Enabled]);
  end
  else
  begin
    SomeThingChanged;
{    GetRecNum;}
{    if Assigned(FOnChange) then FOnChange(Self);}
    if Assigned(FOnButtonChange) then FOnButtonChange(Self,
  [Buttons[nbFirst].Enabled,Buttons[nbPrior].Enabled, Buttons[nbNext].Enabled,
  Buttons[nbLast].Enabled,Buttons[nbInsert].Enabled,Buttons[nbDelete].Enabled,Buttons[nbEdit].Enabled,
  Buttons[nbPost].Enabled,Buttons[nbCancel].Enabled,Buttons[nbRefresh].Enabled,
  Buttons[nbSearch].Enabled,Buttons[nbSearchNext].Enabled,
  Buttons[nbSetBookmark].Enabled,Buttons[nbGotoBookmark].Enabled]);
  end;
end;

procedure TXDBNavigator.SomethingChanged;
var
  UpEnable, DnEnable: Boolean;
  canModify:boolean;
begin
  UpEnable := Enabled and FDataLink.Active and not FDataLink.DataSet.BOF;
  DnEnable := Enabled and FDataLink.Active and not FDataLink.DataSet.EOF;
  CanModify := Enabled and FDataLink.Active and FDataLink.DataSet.CanModify;

  {edit}
  Buttons[nbInsert].Enabled := CanModify;
  Buttons[nbEdit].Enabled := CanModify and not FDataLink.Editing;
  Buttons[nbDelete].Enabled := CanModify;
  Buttons[nbPost].Enabled := CanModify and FDataLink.Editing;
  Buttons[nbCancel].Enabled := CanModify and FDataLink.Editing;
  Buttons[nbRefresh].Enabled := not (FDataLink.DataSet is TQuery)
          and not FDataLink.Editing;
  {edit + data}
  Buttons[nbFirst].Enabled := not FDataLink.Editing and UpEnable;
  Buttons[nbPrior].Enabled := not FDataLink.Editing and UpEnable;
  Buttons[nbNext].Enabled := not FDataLink.Editing and DnEnable;
  Buttons[nbLast].Enabled := not FDataLink.Editing and DnEnable;
  Buttons[nbSearch].Enabled := not FDataLink.Editing and Enabled;
  if XT <> nil then
     Buttons[nbSearchNext].Enabled := XT.CanSearchNext and not FDataLink.Editing
  else
     Buttons[nbSearchNext].Enabled := false;
  Buttons[nbSetBookmark].Enabled := not FDataLink.Editing;
  if XT <> nil then
     Buttons[nbGotoBookmark].Enabled := XT.HasBookmark and not FDataLink.Editing
  else
     Buttons[nbGotoBookmark].Enabled := false;

  {data changed}
  GetRecNum;
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TXDBNavigator.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if not (csLoading in ComponentState) then
    ActiveChanged;
end;

procedure TXDBNavigator.SetDataSource(Value: TDataSource; XT: TDBXT);
var
   old: TDatasource;
begin
  if ((fdefaultDBXT <> nil) and (fdefaultDBXT = XT)) or (fdefaultDBXT = nil) then
  begin
  if FDatalink <> nil then
     old:=FDataLink.DataSource
  else
     old:=nil;

  if (old <> value) then
  begin
  FDataLink.DataSource := Value;
  if not (csLoading in ComponentState) then
    ActiveChanged;
  end;
  end;
end;

function TXDBNavigator.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;


procedure TXDBNavigator.SetdefDBXT(Value: TDBXT);
begin
  FdefaultDBXT := Value;
  FDBXT := Value;
  FDataLink.DataSource := nil;
  if not (csLoading in ComponentState) then
    ActiveChanged;
end;

function TXDBNavigator.GetdefDBXT: TDBXT;
begin
  Result := FdefaultDBXT;
end;


procedure TXDBNavigator.SetDBXT(Value: TDBXT);
begin
  if fdefaultDBXT = nil then
  begin
   FDBXT := Value;
   FDataLink.DataSource := nil;
   if not (csLoading in ComponentState) then
     ActiveChanged;
  end;
end;

function TXDBNavigator.GetDBXT: TDBXT;
begin
  Result := FDBXT;
end;

procedure TXDBNavigator.Loaded;
var
  W, H: Integer;
begin
  inherited Loaded;
  W := Width;
  H := Height;
  AdjustSize (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds (Left, Top, W, H);
  InitHints;
  ActiveChanged;
end;

{TNavButton}

destructor TXNavButton.Destroy;
begin
  if FRepeatTimer <> nil then
    FRepeatTimer.Free;
  inherited Destroy;
end;

procedure TXNavButton.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if nsAllowTimer in FNavStyle then
  begin
    if FRepeatTimer = nil then
      FRepeatTimer := TTimer.Create(Self);

    FRepeatTimer.OnTimer := TimerExpired;
    FRepeatTimer.Interval := InitRepeatPause;
    FRepeatTimer.Enabled  := True;
  end;
end;

procedure TXNavButton.MouseUp(Button: TMouseButton; Shift: TShiftState;
                                  X, Y: Integer);
begin
  inherited MouseUp (Button, Shift, X, Y);
  if FRepeatTimer <> nil then
    FRepeatTimer.Enabled  := False;
end;

procedure TXNavButton.TimerExpired(Sender: TObject);
begin
  FRepeatTimer.Interval := RepeatPause;
  if (FState = bsDown) and MouseCapture then
  begin
    try
      Click;
    except
      FRepeatTimer.Enabled := False;
      raise;
    end;
  end;
end;

procedure TXNavButton.Paint;
var
  R: TRect;
begin
  inherited Paint;
  if (GetFocus = Parent.Handle) and
     (FIndex = TXDBNavigator (Parent).FocusedButton) then
  begin
    R := Bounds(0, 0, Width, Height);
    InflateRect(R, -3, -3);
    if FState = bsDown then
      OffsetRect(R, 1, 1);
    DrawFocusRect(Canvas.Handle, R);
  end;
end;

{ TNavDataLink }

constructor TNavDataLink.Create(ANav: TXDBNavigator);
begin
  inherited Create;
  FNavigator := ANav;
end;

destructor TNavDataLink.Destroy;
begin
  FNavigator := nil;
  inherited Destroy;
end;

procedure TNavDataLink.EditingChanged;
begin
  if FNavigator <> nil then FNavigator.EditingChanged;
{  if FNavigator <> nil then FNavigator.SomethingChanged;}
end;

procedure TNavDataLink.DataSetChanged;
begin
  if FNavigator <> nil then FNavigator.DataChanged;
{  if FNavigator <> nil then FNavigator.SomethingChanged;}
end;

procedure TNavDataLink.ActiveChanged;
begin
  if FNavigator <> nil then FNavigator.ActiveChanged;
end;

end.
